#------------------------------------------------------------------
# $Copyright: Copyright (c) 2010 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/TibcoEMS/TibcoEMS.pm,v 1.7 2010/03/19 08:18:26 lpatil Exp $
#
# $Author: lpatil $
#
# $Date: 2010/03/19 08:18:26 $
# $Revision: 1.7 $
#
# Provides:                            TibcoEMS
# Author:                              lpatil
# crAgent Build:                       1.91 2009/09/07 08:42:27
# crAgent Run Date:                    10 Feb, 2010
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package TibcoEMS;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $HADECRYPT ||= "$CLUS_HOME/sbin/${sPrefix}decrypt";

my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: TibcoEMS.pm,v 1.7 2010/03/19 08:18:26 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define TibcoEMS subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String User,
                              IN String EnvFile,
                              IN String EMSHomeDir,
                              IN String ConfigFile,
                              IN String TibEmsServerUrl,
                              IN Integer SecondLevelMonitor,
                              IN String TibUser,
                              IN String TibPassword,
                              IN String MonitorProgram,
                              IN String StartOptions,
                              IN String TibemsadminOptions );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localUser                     UNIX user name for managing a Tibco Enterprise Message Server
 $localEnvFile                  Absolute path to environment file to source
 $localEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $localConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $localTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $localSecondLevelMonitor       Monitor cycle count to run in-depth monitoring
 $localTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $localTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $localMonitorProgram           Full path to external monitor program
 $localStartOptions             List of startup parameters for tibemsd command
 $localTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localUser,
       $localEnvFile,
       $localEMSHomeDir,
       $localConfigFile,
       $localTibEmsServerUrl,
       $localSecondLevelMonitor,
       $localTibUser,
       $localTibPassword,
       $localMonitorProgram,
       $localStartOptions,
       $localTibemsadminOptions ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  User                     [%s]
  EnvFile                  [%s]
  EMSHomeDir               [%s]
  ConfigFile               [%s]
  TibEmsServerUrl          [%s]
  SecondLevelMonitor       [%s]
  TibUser                  [%s]
  TibPassword              [%s]
  MonitorProgram           [%s]
  StartOptions             [%s]
  TibemsadminOptions       [%s]",

  $sMyName,    $sEntryPointName,     $localUser,     $localEnvFile,
  $localEMSHomeDir,    $localConfigFile,     $localTibEmsServerUrl,
  $localSecondLevelMonitor,    $localTibUser,    $localTibPassword,
  $localMonitorProgram,                         $localStartOptions,
  $localTibemsadminOptions );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName User EnvFile EMSHomeDir ConfigFile TibEmsServerUrl
       SecondLevelMonitor     TibUser     TibPassword     MonitorProgram
       StartOptions TibemsadminOptions );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EMSHomeDir ConfigFile TibEmsServerUrl TibUser TibPassword User );
  my @lsOfflineArgs = qw ( EMSHomeDir ConfigFile TibEmsServerUrl TibUser TibPassword User );
  my @lsMonitorArgs = qw ( EMSHomeDir ConfigFile TibEmsServerUrl User );
  my @lsCleanArgs   = qw ( ConfigFile User );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  

  #-----------------------------------------------
  # Validation for online offline entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #--------------------------------
    # Validate parameter EMSHomeDir..
    #--------------------------------
    if ( ! -d $localEMSHomeDir ) {
      &logError ( 20003,
      "<%s> EMSHomeDir [%s] directory does not exist",
      $sMyName, $localEMSHomeDir );
      $iReturnCode = 0; # Failure..
    }

  }

  #------------------------------------------------
  # Validation for online offline  entry point(s)..
  #------------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline|monitor)$/o ) {
    #--------------------------
    # Validate parameter User..
    #--------------------------
    if ( ! defined getpwnam $localUser ) {
      &logError ( 20004,
      "<%s> User [%s] is not a valid UNIX login",
      $sMyName, $localUser );
      $iReturnCode = 0; # Failure..
    }

  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #--------------------------------
    # Validate parameter ConfigFile..
    #--------------------------------
    if ( ! -e $localConfigFile ) {
      &logError ( 20005,
      "<%s> ConfigFile [%s] file does not exist",
      $sMyName, $localConfigFile );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20006,
      "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

  }

  $iReturnCode ||
  &logWarn ( 20007,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String ConfigFile,
                         IN String EMSHomeDir,
                         IN String EnvFile,
                         IN Integer SecondLevelMonitor,
                         IN String StartOptions,
                         IN String TibEmsServerUrl,
                         IN String TibPassword,
                         IN String TibUser,
                         IN String TibemsadminOptions,
                         IN String User );

 Description : Start logic needed for online entry pt

 Parameters :

 $sConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $sEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $sEnvFile                  Absolute path to environment file to source
 $iSecondLevelMonitor       Monitor cycle count to run in-depth monitoring
 $sStartOptions             List of startup parameters for tibemsd command
 $sTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $sTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $sTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $sTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigFile,
       $sEMSHomeDir,
       $sEnvFile,
       $iSecondLevelMonitor,
       $sStartOptions,
       $sTibEmsServerUrl,
       $sTibPassword,
       $sTibUser,
       $sTibemsadminOptions,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20008, "Subroutine <%s> called with:
  ConfigFile               [%s]
  EMSHomeDir               [%s]
  EnvFile                  [%s]
  SecondLevelMonitor       [%s]
  StartOptions             [%s]
  TibEmsServerUrl          [%s]
  TibPassword              [%s]
  TibUser                  [%s]
  TibemsadminOptions       [%s]
  User                     [%s]",

  $sMyName,      $sConfigFile,       $sEMSHomeDir,       $sEnvFile,
  $iSecondLevelMonitor,     $sStartOptions,      $sTibEmsServerUrl,
  $sTibPassword, $sTibUser, $sTibemsadminOptions, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting TibcoEMS processes..
  #-----------------------------------------
  my $sResourceType   = $ENV { VCS_LOG_AGENT_NAME };
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sEntryPointName = $ENV { VCS_LOG_SCRIPT_NAME };
  my $sTibEMSOutFile  = "/tmp/.VRTS$sResourceType/$sResourceName/$sResourceName" .
                        "_$sEntryPointName.slm.out";
  if ( -e $sTibEMSOutFile ) {
    if ( ! unlink($sTibEMSOutFile) ) {
      &logError ( 20144, "<%s> Could not remove online.slm.out file [%s] ",
      $sMyName, $sTibEMSOutFile );
    }
  }
  $sTibEMSOutFile  = "/tmp/.VRTS$sResourceType/$sResourceName/$sResourceName" .
                     "_monitor.slm.out";
  if ( -e $sTibEMSOutFile ) {
    if ( ! unlink($sTibEMSOutFile) ) {
      &logError ( 20145, "<%s> Could not remove monitor.slm.out file [%s] ",
      $sMyName, $sTibEMSOutFile );
    }
  }

  &logTrace ( 20009,
  "<%s> Killing TibcoEMS processes for this instance", $sMyName );

  if ( TibcoEMS::KillTibcoEMSProcesses ( $sConfigFile,
                                         $sUser ) ) {
    &logTrace ( 20010,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20011,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + TibcoEMS::GetEngineTimeout ( 1 );
  
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------

  $sCommand = "$sEMSHomeDir/tibemsd";
  $sCommandArgs = "-config $sConfigFile $sStartOptions &";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20012,
  "<%s> Starting TibcoEMS instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20013,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    TibcoEMS::ExecLocalCmd ( $sUser, 
                             $sEnvFile, 
                             $sCommand, 
                             $sCommandArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout,
                             $sTibPassword,
                             $sEMSHomeDir ) ) {
    &logTrace ( 20014,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20015,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         TibcoEMS::GetProcState ( \@liPids,
                                  $sConfigFile,
                                  $sUser,
                                  1 ) && (            # Running FFLM..
         $iSecondLevelMonitor  ?                      # If enabled..
         VCSDef_MonitorExit::ONLINE ==                # Online after..
         TibcoEMS::SecondLevelMonitor ( $sEMSHomeDir, # Running SLM..
                                        $sEnvFile,
                                        $sTibEmsServerUrl,
                                        $sTibPassword,
                                        $sTibUser,
                                        $sTibemsadminOptions,
                                        $sUser,
                                        $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20016,
      "<%s> Confirmed start-up of this instance of TibcoEMS server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20017,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String ConfigFile,
                        IN String EMSHomeDir,
                        IN String EnvFile,
                        IN String TibEmsServerUrl,
                        IN String TibPassword,
                        IN String TibUser,
                        IN String TibemsadminOptions,
                        IN String User );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $sEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $sEnvFile                  Absolute path to environment file to source
 $sTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $sTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $sTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $sTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigFile,
       $sEMSHomeDir,
       $sEnvFile,
       $sTibEmsServerUrl,
       $sTibPassword,
       $sTibUser,
       $sTibemsadminOptions,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20018, "Subroutine <%s> called with:
  ConfigFile               [%s]
  EMSHomeDir               [%s]
  EnvFile                  [%s]
  TibEmsServerUrl          [%s]
  TibPassword              [%s]
  TibUser                  [%s]
  TibemsadminOptions       [%s]
  User                     [%s]",

  $sMyName,      $sConfigFile,       $sEMSHomeDir,       $sEnvFile,
  $sTibEmsServerUrl,           $sTibPassword,            $sTibUser,
  $sTibemsadminOptions, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + TibcoEMS::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  $sCommand = "$sEMSHomeDir/tibemsadmin";
  $sCommandArgs = "-server $sTibEmsServerUrl -user $sTibUser $sTibemsadminOptions << EOF
$sTibPassword
shutdown
yes
exit
EOF";
  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20019,
  "<%s> Stopping TibcoEMS with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20020,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          TibcoEMS::ExecLocalCmd ( $sUser, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout,
                                   $sTibPassword ) ) {
    &logTrace ( 20021,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }


  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20022,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      TibcoEMS::GetProcState ( \@liPids,
                             $sConfigFile,
                               $sUser,
                               0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20023,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20024,
        "<%s> Confirmed shutdown of TibcoEMS instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting TibcoEMS processes..
  #-----------------------------------------
  &logTrace ( 20025,
  "<%s> Killing TibcoEMS processes for this instance", $sMyName );

  if ( TibcoEMS::KillTibcoEMSProcesses ( $sConfigFile,
                                         $sUser ) ) {
    &logTrace ( 20026,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20027,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String ConfigFile,
                        IN String EMSHomeDir,
                        IN String EnvFile,
                        IN String TibEmsServerUrl,
                        IN String TibPassword,
                        IN String TibUser,
                        IN String TibemsadminOptions,
                        IN String User );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $sEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $sEnvFile                  Absolute path to environment file to source
 $sTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $sTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $sTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $sTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sConfigFile,
       $sEMSHomeDir,
       $sEnvFile,
       $sTibEmsServerUrl,
       $sTibPassword,
       $sTibUser,
       $sTibemsadminOptions,
       $sUser ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20028, "Subroutine <%s> called with:
  CleanReason              [%s]
  ConfigFile               [%s]
  EMSHomeDir               [%s]
  EnvFile                  [%s]
  TibEmsServerUrl          [%s]
  TibPassword              [%s]
  TibUser                  [%s]
  TibemsadminOptions       [%s]
  User                     [%s]",

  $sMyName, $sCleanReason, $sConfigFile,  $sEMSHomeDir,  $sEnvFile,
  $sTibEmsServerUrl,           $sTibPassword,            $sTibUser,
  $sTibemsadminOptions, $sUser );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + TibcoEMS::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sEMSHomeDir/tibemsadmin";
  $sCommandArgs = "-server $sTibEmsServerUrl -user $sTibUser $sTibemsadminOptions << EOF
$sTibPassword
shutdown
yes
exit
EOF";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20029,
    "<%s> Attempting to stop TibcoEMS with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20030,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( TibcoEMS::ExecLocalCmd ( $sUser, 
                                     $sEnvFile, 
                                     $sCommand, 
                                     $sCommandArgs, 
                                     \$iExitCode, 
                                     $iTimeout, 
                                     \@lsCmdout,
                                     $sTibPassword ) ) {

      &logTrace ( 20031, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #-----------------------------------------
  # Whack any persisting TibcoEMS processes..
  #-----------------------------------------
  &logTrace ( 20032,
  "<%s> Killing TibcoEMS processes for this instance", $sMyName );

  if ( TibcoEMS::KillTibcoEMSProcesses ( $sConfigFile,
                                         $sUser ) ) {
    &logTrace ( 20033,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20034,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String ConfigFile,
                                   IN String User );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigFile,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20035, "Subroutine <%s> called:
  ConfigFile               [%s]
  User                     [%s]",

  $sMyName, $sConfigFile, $sUser );
  

  my @liDummyPids = ();
  my $iReturnCode = TibcoEMS::GetProcState ( \@liDummyPids,
                                           $sConfigFile,
                                             $sUser,
                                             1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20036,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20037,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20038,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String ConfigFile,
                               IN String User,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sConfigFile,
       $sUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20039, "Subroutine <%s> called with:
  OnlinePids               [%s]
  ConfigFile               [%s]
  User                     [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sConfigFile, $sUser, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20040, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( TibcoEMS::GetSrvrDetails ( $sConfigFile, 
                                    \%hsSrvrDetails ) ) {
      &logTrace ( 20041,
      "<%s> Got details for monitoring TibcoEMS instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20042,
      "<%s> Failed to get details for monitoring TibcoEMS instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nTibcoEMS process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = TibcoEMS::FFLMStatus ( $sPidFile,
                                         $sProc,
                                         $iUid,
                                         \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20043,
      "<%s> TibcoEMS instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20044,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20045,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20046,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20047,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20048, "<%s> Searching PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      #-----------------------------------------------
      # Look for processes that were offline in FFLM..
      #-----------------------------------------------
      if ( $iUseFFLM != 2 ) { # Request to filter..
        foreach $sKey ( keys %hsCopy ) {
          @lsUpdates = (); # Reset..
          @lsValues = @{ $hsCopy { $sKey } }; # Get values..
  
          #--------------------------------------------------
          # Should have details per three elements of array..
          # <PIDFILE> <PROCESS> <STATE>
          #--------------------------------------------------
          while ( ( $sPidFile, $sProc, $iFound ) =
                    splice ( @lsValues, 0, 3 ) ) {
            if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
              if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                unshift ( @$rliOnlinePids, $iPid );         # Store..
  
                &logTrace ( 20049,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                $sProc );

                $hiProcPids { $sKey } = $iPid;        # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
              }
            }
  
            #----------------
            # Update status..
            #----------------
            push ( @lsUpdates, $sPidFile, $sProc, $iFound );
          }
  
          @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
        }
        }
        else { # Request for kill..
          &logTrace ( 20050,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20051,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( TibcoEMS::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20052,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20053,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20054,
    "<%s> Found [%s] PIDs [%s] for this instance of TibcoEMS.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20055,
    "<%s> TibcoEMS process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String EMSHomeDir,
                          IN String EnvFile,
                          IN String MonitorProgram,
                          IN Integer SecondLevelMonitor,
                          IN String TibEmsServerUrl,
                          IN String TibPassword,
                          IN String TibUser,
                          IN String TibemsadminOptions,
                          IN String User );

 Description: Performs the in-depth monitoring

 Parameters :

 $sEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $sEnvFile                  Absolute path to environment file to source
 $sMonitorProgram           Full path to external monitor program
 $iSecondLevelMonitor       Monitor cycle count to run in-depth monitoring
 $sTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $sTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $sTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $sTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEMSHomeDir,
       $sEnvFile,
       $sMonitorProgram,
       $iSecondLevelMonitor,
       $sTibEmsServerUrl,
       $sTibPassword,
       $sTibUser,
       $sTibemsadminOptions,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20056, "Subroutine <%s> called with:
  EMSHomeDir               [%s]
  EnvFile                  [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]
  TibEmsServerUrl          [%s]
  TibPassword              [%s]
  TibUser                  [%s]
  TibemsadminOptions       [%s]
  User                     [%s]",

  $sMyName,     $sEMSHomeDir,     $sEnvFile,      $sMonitorProgram,
  $iSecondLevelMonitor,      $sTibEmsServerUrl,      $sTibPassword,
  $sTibUser, $sTibemsadminOptions, $sUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         TibcoEMS::DelayedAttributeCheck ( $sEMSHomeDir,
                                           $sEnvFile,
                                           $sTibEmsServerUrl,
                                           $sTibPassword,
                                           $sTibUser,
                                           $sTibemsadminOptions,
                                           $sUser ) ) ) {
      &logTrace ( 20057,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20058,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20059,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + TibcoEMS::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         TibcoEMS::SecondLevelMonitor ( $sEMSHomeDir,
                                        $sEnvFile,
                                        $sTibEmsServerUrl,
                                        $sTibPassword,
                                        $sTibUser,
                                        $sTibemsadminOptions,
                                        $sUser,
                                        $iTillWhen ) ) ) {

      &logTrace ( 20060,
      "<%s> TibcoEMS instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20061,
      "<%s> TibcoEMS instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + TibcoEMS::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         TibcoEMS::ExternalMonitor ( $sUser,
                                     $sEnvFile,
                                     $sExtCmd,
                                     $sExtCmdArgs,
                                     $iTillWhen ) ) ) {
      &logTrace ( 20062,
      "<%s> TibcoEMS instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20063,
      "<%s> TibcoEMS instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String EMSHomeDir,
                                        IN String EnvFile,
                                        IN String TibEmsServerUrl,
                                        IN String TibPassword,
                                        IN String TibUser,
                                        IN String TibemsadminOptions,
                                        IN String User );

 Description : Just In Time Validation

 Parameters :

 $sEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $sEnvFile                  Absolute path to environment file to source
 $sTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $sTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $sTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $sTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEMSHomeDir,
       $sEnvFile,
       $sTibEmsServerUrl,
       $sTibPassword,
       $sTibUser,
       $sTibemsadminOptions,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20064, "Subroutine <%s> called with:
  EMSHomeDir               [%s]
  EnvFile                  [%s]
  TibEmsServerUrl          [%s]
  TibPassword              [%s]
  TibUser                  [%s]
  TibemsadminOptions       [%s]
  User                     [%s]",

  $sMyName,     $sEMSHomeDir,     $sEnvFile,     $sTibEmsServerUrl,
  $sTibPassword, $sTibUser, $sTibemsadminOptions, $sUser );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( EMSHomeDir   EnvFile    TibEmsServerUrl    TibPassword    TibUser
TibemsadminOptions User );
  my @lsReqdArgs = qw ( EMSHomeDir   TibEmsServerUrl    TibPassword    TibUser User );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20065,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #---------------------------------
  # Validate parameter EMSHomeDir..
  #---------------------------------
  if ( ! -d $sEMSHomeDir ) {
    &logError ( 20066,
    "<%s> EMSHomeDir [%s] directory does not exist",
    $sMyName, $sEMSHomeDir );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #---------------------------
  # Validate parameter User..
  #---------------------------
  if ( ! defined getpwnam $sUser ) {
    &logError ( 20067,
    "<%s> User [%s] is not a valid UNIX login",
    $sMyName, $sUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String EMSHomeDir,
                                     IN String EnvFile,
                                     IN String TibEmsServerUrl,
                                     IN String TibPassword,
                                     IN String TibUser,
                                     IN String TibemsadminOptions,
                                     IN String User,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sEMSHomeDir               Full path of directory in which the tibemsd binary file and tibemsadmin utility is located
 $sEnvFile                  Absolute path to environment file to source
 $sTibEmsServerUrl          Tibco EMS server URL used to specify the -server parameter for tibemsadmin utility
 $sTibPassword              Password of the Tibco user to connect to server using tibemsadmin utility
 $sTibUser                  Tibco user name to connect to server using tibemsadmin utility
 $sTibemsadminOptions       List of startup parameters for tibemsadmin command, used during in-depth monitoring
 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sEMSHomeDir,
       $sEnvFile,
       $sTibEmsServerUrl,
       $sTibPassword,
       $sTibUser,
       $sTibemsadminOptions,
       $sUser,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20068, "Subroutine <%s> called with:
  EMSHomeDir               [%s]
  EnvFile                  [%s]
  TibEmsServerUrl          [%s]
  TibPassword              [%s]
  TibUser                  [%s]
  TibemsadminOptions       [%s]
  User                     [%s]
  TillWhen                 [%s]",

  $sMyName,     $sEMSHomeDir,     $sEnvFile,     $sTibEmsServerUrl,
  $sTibPassword,    $sTibUser,    $sTibemsadminOptions,     $sUser,
  $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  my $sResourceType   = $ENV { VCS_LOG_AGENT_NAME };
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  my $sEntryPointName = $ENV { VCS_LOG_SCRIPT_NAME };
  my $sTibEMSOutFile  = "/tmp/.VRTS$sResourceType/$sResourceName/$sResourceName" .
                        "_$sEntryPointName.slm.out";

  #---------------------------
  # Re-direct command output..
  #---------------------------
  my $sRdStr = ( getpwnam $sUser ) [-1] =~ /csh$/o ?">& $sTibEMSOutFile" :
                                                    "1> $sTibEMSOutFile 2>&1";

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sEMSHomeDir/tibemsadmin";
  $sCommandArgs = "-server $sTibEmsServerUrl -user " .
                  "$sTibUser $sTibemsadminOptions $sRdStr << EOF 
$sTibPassword
exit
EOF";

  #----------------------------------
  # Set LANG for better performance..
  #----------------------------------
  my $sLang = $ENV { LANG }; # Backup..
  $ENV { LANG } = 'C';    # Now set..

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20069,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( TibcoEMS::ExecLocalCmd ( $sUser, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout,
                                   $sTibPassword ) ) {

    &logTrace ( 20070,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #----------
  # Restore..
  #----------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String User,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sUser                     UNIX user name for managing a Tibco Enterprise Message Server
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20073, "Subroutine <%s> called with:
  User                     [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20074,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20075,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20076,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20077,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20078,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  TibcoEMS::ExecLocalCmd ( $sUser, 
                           $sEnvFile, 
                           $sCmd, 
                           $sCmdArgs, 
                           \$iExitCode, 
                           $iTimeout, 
                           \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20079,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20080,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20081, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20082, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20083,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20084,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20085,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20086, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20087, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20088, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20089, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20090,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20091,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20092,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( TibcoEMS::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20093,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20094,
          "<%s> PID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20095,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20096,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20097, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillTibcoEMSProcesses ( IN String ConfigFile,
                                         IN String User );

 Description  : Kills processes for this instance of TibcoEMS
 
 Parameters :

 $sConfigFile             Full path and file name of the main configuration file for Tibco EMS
 $sUser                   UNIX user name for managing a Tibco Enterprise Message Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillTibcoEMSProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigFile,
       $sUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20098, "Subroutine <%s> called with:
  ConfigFile               [%s]
  User                     [%s]",

  $sMyName, $sConfigFile, $sUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'TibcoEMS::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    TibcoEMS::GetProcState ( \@liGoners,
                           $sConfigFile,
                             $sUser,
                             0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20099,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20100,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20101,
    "<%s> Could not locate any process this instance of TibcoEMS",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sHideString,
       $sEMSHomeDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20102, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]
  EMSHomeDir  [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout, $sEMSHomeDir );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20103,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {                             # Get decrypted..
                          s/([^\w-])/\\$1/g || 1          # Escape non-words.. 
                        } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20104, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20105,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20106,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20107,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20108,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );
  $sEMSHomeDir ||= ''; # Prevent undefs..

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           $sEMSHomeDir,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20109,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20110,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20111,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20112,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20113, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20114,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  goto $sSwitch; {
    solaris:

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;

    hpux:

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20115, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20116,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20117,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String ConfigFile,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sConfigFile               Full path and file name of the main configuration file for Tibco EMS
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sConfigFile,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20118, "Subroutine <%s> called with:
  ConfigFile               [%s]
  SrvrDetails              [%s]",

  $sMyName, $sConfigFile, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( qr@tibemsd\s.*-config\s+$sConfigFile\b@ );

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  my $sOldMask = umask ( 000 );

  if ( -d $sPIDDir ) {
    &logTrace ( 20119, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0777 ) ) {
    &logWarn ( 20120, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  umask ( $sOldMask );
  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20121, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20122,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20123,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20124, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20125, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "TibcoEMS" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20126,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20127, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20128, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20129,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

1; # So PERL knows its included..
